{-|
Module      : Gargantext.API.Metrics
Description : Server API
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

Metrics API

-}

{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeOperators   #-}

module Gargantext.API.Metrics
    where

import Control.Lens ((%~))
import Control.Lens.Getter (Getting)
import Control.Lens.Setter (ASetter)
import Data.HashMap.Strict qualified as HashMap
import Data.Vector (Vector)
import Gargantext.API.HashedResponse (HashedResponse, constructHashedResponse, hash)
import Gargantext.API.Ngrams.NgramsTree (NgramsTree)
import Gargantext.API.Ngrams.Types (TabType, ngramsTypeFromTabType, unNgramsTerm)
import Gargantext.API.Prelude (IsGargServer)
import Gargantext.API.Routes.Named.Metrics qualified as Named
import Gargantext.Core.NodeStory (HasNodeStory)
import Gargantext.Core.Text.Metrics (Scored(..), {-normalizeGlobal,-} normalizeLocal)
import Gargantext.Core.Types (CorpusId, ListId, ListType(..))
import Gargantext.Core.Types.Query (Limit)
import Gargantext.Core.Viz.Chart (chartData, histoData, treeData)
import Gargantext.Core.Viz.Types (Histo)
import Gargantext.Database.Action.Metrics qualified as Metrics
import Gargantext.Database.Admin.Types.Hyperdata (HyperdataList(..), hl_chart, hl_pie, hl_scatter, hl_tree)
import Gargantext.Database.Admin.Types.Metrics (ChartMetrics(..), Metric(..), Metrics(..))
import Gargantext.Database.Admin.Types.Node (NodeId)
import Gargantext.Database.Prelude (DBCmd)
import Gargantext.Database.Query.Table.Node (defaultList, getNodeWith)
import Gargantext.Database.Query.Table.Node.Error (HasNodeError)
import Gargantext.Database.Query.Table.Node.UpdateOpaleye (updateHyperdata)
import Gargantext.Database.Schema.Node (node_hyperdata)
import Gargantext.Prelude hiding (hash)
import Gargantext.System.Logging (logLocM, LogLevel(DEBUG), MonadLogger)
import Servant.Server.Generic (AsServerT)


-------------------------------------------------------------
scatterApi :: IsGargServer err env m => NodeId -> Named.ScatterAPI (AsServerT m)
scatterApi id' = Named.ScatterAPI
  { sepGenEp        = getScatter id'
  , scatterUpdateEp = updateScatter id'
  , scatterHashEp   = getScatterHash id'
  }

getScatter :: HasNodeStory env err m
           => CorpusId
           -> Maybe ListId
           -> TabType
           -> m (HashedResponse Metrics)
getScatter cId maybeListId tabType = do
  metricsGetter cId maybeListId tabType hl_scatter (\cId' ml tt -> updateScatter' cId' ml tt Nothing)

updateScatter :: HasNodeStory env err m
              => CorpusId
              -> Maybe ListId
              -> TabType
              -> Maybe Limit
              -> m ()
updateScatter cId mListId tabType maybeLimit = do
  listId <- getListOrDefault cId mListId
  -- printDebug "[updateScatter] cId" cId
  -- printDebug "[updateScatter] maybeListId" maybeListId
  -- printDebug "[updateScatter] tabType" tabType
  -- printDebug "[updateScatter] maybeLimit" maybeLimit
  _ <- updateScatter' cId listId tabType maybeLimit
  pure ()

updateScatter' :: HasNodeStory env err m
               => CorpusId
               -> ListId
               -> TabType
               -> Maybe Limit
               -> m Metrics
updateScatter' cId listId tabType maybeLimit = do
  (ngs', scores) <- Metrics.getMetrics cId listId tabType maybeLimit

  let
    metrics      = fmap (\(Scored t s1 s2) -> Metric { m_label = unNgramsTerm t
                                                     , m_x = s1
                                                     , m_y = s2
                                                     , m_cat = listType t ngs' })
                 $ fmap normalizeLocal scores
    listType t m = maybe (panicTrace errorMsg) fst $ HashMap.lookup t m
    errorMsg     = "API.Node.metrics: key absent"

  updateNodeMetrics listId tabType hl_scatter (Metrics metrics)

getScatterHash :: HasNodeStory env err m
               => CorpusId
               -> Maybe ListId
               -> TabType
               -> m Text
getScatterHash cId maybeListId tabType = do
  hash <$> getScatter cId maybeListId tabType


-------------------------------------------------------------
chartApi :: IsGargServer err env m => NodeId -> Named.ChartAPI (AsServerT m)
chartApi id' = Named.ChartAPI
  { getChartEp = getChart id'
  , updateChartEp = updateChart id'
  , chartHashEp = getChartHash id'
  }

-- TODO add start / end
getChart :: HasNodeStory env err m
         => CorpusId
         -> Maybe ListId
         -> TabType
         -> m (HashedResponse (ChartMetrics Histo))
getChart cId maybeListId tabType = do
  metricsGetter cId maybeListId tabType hl_chart updateChart'
  
updateChart :: (HasNodeStory env err m, MonadLogger m)
            => CorpusId
            -> Maybe ListId
            -> TabType
            -> m ()
updateChart cId mListId tabType = do
  listId <- getListOrDefault cId mListId
  $(logLocM) DEBUG $ "[updateChart] cId: " <> show cId
  $(logLocM) DEBUG $ "[updateChart] listId: " <> show listId
  $(logLocM) DEBUG $ "[updateChart] tabType: " <> show tabType
  _ <- updateChart' cId listId tabType
  pure ()

updateChart' :: HasNodeError err
             => CorpusId
             -> ListId
             -> TabType
             -> DBCmd err (ChartMetrics Histo)
updateChart' cId listId tabType = do
  metrics <- histoData cId
  updateNodeMetrics listId tabType hl_chart (ChartMetrics metrics)


getChartHash :: (HasNodeStory env err m, MonadLogger m)
             => CorpusId
             -> Maybe ListId
             -> TabType
             -> m Text
getChartHash cId maybeListId tabType = do
  hash <$> getChart cId maybeListId tabType

-------------------------------------------------------------

pieApi :: IsGargServer err env m => NodeId -> Named.PieAPI (AsServerT m)
pieApi id' = Named.PieAPI
  { getPieChartEp    = getPie id'
  , pieChartUpdateEp = updatePie id'
  , pieHashEp        = getPieHash id'
  }

getPie :: HasNodeStory env err m
       => CorpusId
       -> Maybe ListId
       -> TabType
       -> m (HashedResponse (ChartMetrics Histo))
getPie cId maybeListId tabType = do
  metricsGetter cId maybeListId tabType hl_pie updatePie'

updatePie :: (HasNodeStory env err m, MonadLogger m)
          => CorpusId
          -> Maybe ListId
          -> TabType
          -> m ()
updatePie cId mListId tabType = do
  listId <- getListOrDefault cId mListId
  $(logLocM) DEBUG $ "[updatePie] cId: " <> show cId
  $(logLocM) DEBUG $ "[updatePie] mListId: " <> show mListId
  $(logLocM) DEBUG $ "[updatePie] tabType: " <> show tabType
  _ <- updatePie' cId listId tabType
  pure ()

updatePie' :: (HasNodeStory env err m)
           => CorpusId
           -> ListId
           -> TabType
           -> m (ChartMetrics Histo)
updatePie' cId listId tabType = do
  metrics <- chartData cId (ngramsTypeFromTabType tabType) MapTerm
  updateNodeMetrics listId tabType hl_pie (ChartMetrics metrics)

getPieHash :: HasNodeStory env err m
           => CorpusId
           -> Maybe ListId
           -> TabType
           -> m Text
getPieHash cId maybeListId tabType = do
  hash <$> getPie cId maybeListId tabType

-------------------------------------------------------------
-- | Tree metrics API
treeApi :: IsGargServer err env m => NodeId -> Named.TreeAPI (AsServerT m)
treeApi id' = Named.TreeAPI
  { treeChartEp = getTree id'
  , treeChartUpdateEp = updateTree id'
  , treeHashEp = getTreeHash id'
  }

getTree :: HasNodeStory env err m
        => CorpusId
        -> Maybe ListId
        -> TabType
        -> ListType
        -> m (HashedResponse (ChartMetrics (Vector NgramsTree)))
getTree cId maybeListId tabType listType = do
  metricsGetter cId maybeListId tabType hl_tree (\cId' l tt -> updateTree' cId' l tt listType)

updateTree :: (HasNodeStory env err m, MonadLogger m)
           => CorpusId
           -> Maybe ListId
           -> TabType
           -> ListType
           -> m ()
updateTree cId mListId tabType listType = do
  listId <- getListOrDefault cId mListId
  $(logLocM) DEBUG $ "[updateTree] cId: " <> show cId
  $(logLocM) DEBUG $ "[updateTree] mListId: " <> show mListId
  $(logLocM) DEBUG $ "[updateTree] tabType: " <> show tabType
  $(logLocM) DEBUG $ "[updateTree] listType: " <> show listType
  _ <- updateTree' cId listId tabType listType
  pure ()

updateTree' :: HasNodeStory env err m
            => CorpusId
            -> ListId
            -> TabType
            -> ListType
            -> m (ChartMetrics (Vector NgramsTree))
updateTree' cId listId tabType listType = do
  metrics <- treeData cId (ngramsTypeFromTabType tabType) listType
  updateNodeMetrics listId tabType hl_tree (ChartMetrics metrics)

getTreeHash :: HasNodeStory env err m
            => CorpusId
            -> Maybe ListId
            -> TabType
            -> ListType
            -> m Text
getTreeHash cId maybeListId tabType listType = do
  hash <$> getTree cId maybeListId tabType listType


--------

metricsGetter :: (HasNodeStory env err m, ToJSON a)
              => CorpusId
              -> Maybe ListId
              -> TabType
              -> Getting (HashMap.HashMap TabType a) HyperdataList (HashMap.HashMap TabType a)
              -> (CorpusId -> ListId -> TabType -> m a)
              -> m (HashedResponse a)
metricsGetter cId mListId tabType l up = do
  listId <- getListOrDefault cId mListId
  node <- getNodeWith listId (Proxy :: Proxy HyperdataList)
  let metricsMap = node ^. node_hyperdata ^. l
      mMetrics = HashMap.lookup tabType metricsMap

  metrics <- case mMetrics of
    Just m -> pure m
    Nothing    -> do
      up cId listId tabType

  pure $ constructHashedResponse metrics

getListOrDefault :: HasNodeError err => CorpusId -> Maybe ListId -> DBCmd err ListId
getListOrDefault cId mListId = case mListId of
    Just lid -> pure lid
    Nothing  -> defaultList cId



updateNodeMetrics :: (HasNodeError err)
                  => ListId
                  -> TabType
                  -- -> _Lens
                  -> ASetter HyperdataList HyperdataList (HashMap.HashMap TabType a) (HashMap.HashMap TabType a)
                  -> a
                  -> DBCmd err a
updateNodeMetrics listId tabType setter metrics = do
  node <- getNodeWith listId (Proxy :: Proxy HyperdataList)
  let hl = node ^. node_hyperdata
  _ <- updateHyperdata listId (setter %~ (HashMap.insert tabType metrics) $ hl)

  pure metrics
